//  Project       : Toolbox
//  Workpackage   : libCommExtraction 
//
//               Copyright (C) 2023, ...
//            Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
//                       German Aerospace Center
//
//       Institut fuer Flugfuehrung/Institute of Flight Guidance
//              Tel. +49 531 295 2501, Fax: +49 531 295 2550
//                 WWW: http://www.bs.dlr.de/FL/
//
//  These coded instructions, statements, and computer programs contain
//  unpublished proprietary information of the German Aerospace Center
//  and are protected by the copyright law. They may not be disclosed to
//  third parties or copied or duplicated in any form,in whole or in part,
//  without the prior written consent of the German Aerospace Center


/*!
\file  main.cxx

\author Hartmut Helmke (HHe)

\date 2023-07-08 HHe : first version

*/

#include <iostream>
#include <string>
#include <sstream>
#include <iomanip>
using namespace std;

#include "dosFarben.h"
#include "testMakros.h"

#include "googleTestEmulation.h"

#   define LOG_METHOD(a,b,c,d) 
//#   define TRACE(a) /* comment line out, if you want output by macro TRACE */
#ifndef TRACE
#   include <fstream>
std::fstream g_file("tracefile.trc", ios::out);
#   define TRACE(a) g_file << a << endl;
#endif

#include "C_String.h"
#include "StringUtil.h" // output operator << for vector<string>
using namespace libutil;


// We calculate the Levenshtein distance of Tier and Tor which should be 2
bool LevenshteinDistTierTor();
bool LevenshteinDistWords();
bool LevenshteinDistTierTiere();
bool LevenshteinDistRenTierTiere();
bool CallsignTest();
bool ThalesExample();
bool ThalesExample2();
bool ThalesExample3();
bool differentDistCalc();
bool differentDistCalc2();
bool differentDistCalc3();
bool differentInsSubstDelValues();
bool CreateInsDelSubsAsString();
bool TestLevenshteinDistLong2Helmke();
bool PrettyPrintTest();



class Levenshtein
{
public:
   // astr1
   Levenshtein(
      const std::vector<std::string>& astr1,  // first string compared to astr2
      const std::vector<std::string>& astr2)  // i.e. LD between astr1 and astr2 is calculated
      : mstr1(astr1), mstr2(astr2),
      mb_backtrackCalculated(false)
   {
      ;
   }
   virtual ~Levenshtein() {  }

   //! returning the Levenshtein distance between astr1 and astr2
   int CalcLevenshteinDistance();
   std::string GetMatrixAsString() const;
   std::string backtrace() const;

   int GetSubstCnt() const {
      if (!mb_backtrackCalculated) CalcSubsDelIns();
      return mi_subsCnt;
   }
   int GetInsCnt() const {
      if (!mb_backtrackCalculated) CalcSubsDelIns();
      return mi_insCnt;
   }
   int GetDelCnt() const {
      if (!mb_backtrackCalculated) CalcSubsDelIns();
      return mi_delCnt;
   }
   std::string GetPrettyPrint(int ai_ld, std::string astr_goldText = "",
      std::string astr_recognText = "") const;

private:
   std::string backtraceInternal(int st1Ind, int st2Ind, int& stepCnt) const;
   void backtraceInternalMember(int st1Ind, int st2Ind) const;
   bool Check(int st1Ind, int st2Ind) const {
      return (st1Ind < mi_zCnt_m1&& st2Ind < mi_spCnt_n2&&
         st1Ind* mi_spCnt_n2 + st2Ind < mi_zCnt_m1* mi_spCnt_n2);
   }
   // first index runs over size of mstr1
   int Get(int st1Ind, int st2Ind) const {
      return (
         //Check(st1Ind, st2Ind) ? mpi_mat[st1Ind * mi_spCnt_n2 + st2Ind] : -100
         mi_mat[st1Ind * mi_spCnt_n2 + st2Ind]);
   }
   int& GetMod(int st1Ind, int st2Ind) {
      return (
         //Check(st1Ind, st2Ind) ? mpi_mat[st1Ind * mi_spCnt_n2 + st2Ind] : mpi_mat[0]
         mi_mat[st1Ind * mi_spCnt_n2 + st2Ind]
         );
   }
   void CalcSubsDelIns()const;

   std::vector<std::string> mstr1;
   std::vector<std::string> mstr2;
   int mi_zCnt_m1;   // size of mstr1
   int mi_spCnt_n2;  // size of mstr2

   enum { MAX_WORD_NUMBER = 25, 
      MAX_ARRAY_SIZE = (MAX_WORD_NUMBER + 1) * (MAX_WORD_NUMBER + 1)
   };
   // contains the matrix, used for LD calculation as a vector
   int mi_mat[MAX_ARRAY_SIZE];
   mutable bool mb_backtrackCalculated;
   mutable int mi_subsCnt;
   mutable int mi_insCnt;
   mutable int mi_delCnt;
};



bool GoldAndRecogAsPrettyString(
   const std::vector<std::string>& goldVec,
   const std::vector<std::string>& recognVec,
   const std::vector<std::string>& opsVec,
   std::string& prettyGold,
   std::string& prettyRecogn);




int main()
{
	bool result = true;

   PERFORM_AND_OUTPUT(LevenshteinDistTierTor);
   PERFORM_AND_OUTPUT(LevenshteinDistWords);
   PERFORM_AND_OUTPUT(LevenshteinDistTierTiere);
   PERFORM_AND_OUTPUT(LevenshteinDistRenTierTiere);
   PERFORM_AND_OUTPUT(CallsignTest);
   PERFORM_AND_OUTPUT(ThalesExample);
   PERFORM_AND_OUTPUT(ThalesExample2);
   PERFORM_AND_OUTPUT(ThalesExample3);
   PERFORM_AND_OUTPUT(differentDistCalc);
   PERFORM_AND_OUTPUT(differentDistCalc2);
   PERFORM_AND_OUTPUT(differentDistCalc3);
   PERFORM_AND_OUTPUT(differentInsSubstDelValues);
   PERFORM_AND_OUTPUT(CreateInsDelSubsAsString);
   PERFORM_AND_OUTPUT(TestLevenshteinDistLong2Helmke);
   PERFORM_AND_OUTPUT(PrettyPrintTest);

   

   if (true == result)
   {
      printScreenColorOnceVal(cout, GREEN_SCREEN_COLOR, "Alle Tests erfolgreich\n");
   }
   else
   {
      printScreenColorOnceVal(cout, RED_SCREEN_COLOR, " Fehler Tests aufgetreten! ***\n");
   }



}





   static string StringVecToOneString(const vector<string>& v)
   {
      string blank = "";
      string str_retValue;
      auto iter = v.begin();
      while (iter != v.end())
      {
         str_retValue += blank + *iter;
         blank = " ";
         ++iter;
      }
      return str_retValue;
   }


   // We calculate the Levenshtein distance of two vectors
   TEST(LevenshteinTest, LevenshteinDistWords)
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "LevenshteinDistWords", "");

      vector<string> s1{ "Tango", "ind", "echo", "romeo" };
      vector<string> s2{ "Ind", "echo", "romeo", "abba" };
      Levenshtein dist(s1, s2);
      auto ld = dist.CalcLevenshteinDistance();
      EXPECT_EQ(3, ld);
      TRACE("Needed steps: " << dist.backtrace());
      TRACE(dist.GetPrettyPrint(ld, "s1", "and s2"));
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }



   /* We check whether PrettyPrint provides the expected output
   * The test does not consider the different of \n on LINUX and windows
   */
   TEST(LevenshteinTest, PrettyPrintTest)
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "LevenshteinDistWords", "");

      vector<string> s1{ "Tango", "ind" };
      vector<string> s2{ "Ind", "echo" };
      Levenshtein dist(s1, s2);
      auto ld = dist.CalcLevenshteinDistance();
      string resString = dist.GetPrettyPrint(ld, "s1", "and s2");
      TRACE("result String:" << resString);
      string expString = "s1     : Tango ind  \n"
         "and s2 : Ind   echo  // LD: 2, subs: 2 / ins : 0 / del : 0 gold words: 2\n";
      TRACE("expected String:" << expString);
      EXPECT_EQ(expString, resString);
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }

   // We calculate the Levenshtein distance of Tier and Tor which should be 2
   TEST(LevenshteinTest, LevenshteinDistTierTor)
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "LevenshteinDistTierTor", "");

      vector<string> s1{ "T", "i", "e", "r" };
      vector<string> s2{ "T", "o", "r" };
      Levenshtein dist(s1, s2);
      EXPECT_EQ(2, dist.CalcLevenshteinDistance());
      TRACE("Levenshtein Matrix of Tier/Tor\n" << dist.GetMatrixAsString());
      TRACE("Needed steps: " << dist.backtrace());
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }


   TEST(LevenshteinTest, LevenshteinDistTierTiere)
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "LevenshteinDistTierTiere", "");

      vector<string> s1{ "T", "i", "e", "r" };
      vector<string> s2{ "T", "i", "e", "r", "e" };
      Levenshtein dist2(s2, s1);
      EXPECT_EQ(1, dist2.CalcLevenshteinDistance());
      TRACE("Levenshtein Matrix of Tiere/Tier\n" << dist2.GetMatrixAsString());
      TRACE("Needed steps: " << dist2.backtrace());
      EXPECT_EQ(0, dist2.GetSubstCnt());
      EXPECT_EQ(1, dist2.GetDelCnt());
      EXPECT_EQ(0, dist2.GetInsCnt());

      Levenshtein dist(s1, s2);
      EXPECT_EQ(1, dist.CalcLevenshteinDistance());
      TRACE("Levenshtein Matrix of Tier/Tiere\n" << dist.GetMatrixAsString());
      TRACE("Needed steps: " << dist.backtrace());
      EXPECT_EQ(0, dist.GetSubstCnt());
      EXPECT_EQ(0, dist.GetDelCnt());
      // one Insertion into Tier results in Tiere
      EXPECT_EQ(1, dist.GetInsCnt());

      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }

   TEST(LevenshteinTest, LevenshteinDistRenTierTiere)
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "LevenshteinDistRenTierTiere", "");
      vector<string> s1{ "R", "e", "n", "T", "i", "e", "r" };
      vector<string> s2{ "T", "i", "e", "r", "e" };
      {
         Levenshtein dist(s1, s2);
         EXPECT_EQ(4, dist.CalcLevenshteinDistance());
         TRACE("Levenshtein Matrix of RenTier/Tiere\n" << dist.GetMatrixAsString());
         TRACE("Needed steps: " << dist.backtrace());
         TRACE("subs: " << dist.GetSubstCnt() << ", ins :" << dist.GetInsCnt()
            << " del: " << dist.GetDelCnt());
         EXPECT_EQ(0, dist.GetSubstCnt());
         EXPECT_EQ(3, dist.GetDelCnt());
         EXPECT_EQ(1, dist.GetInsCnt());
      }
      {
         Levenshtein dist2(s2, s1);
         EXPECT_EQ(4, dist2.CalcLevenshteinDistance());
         TRACE("Levenshtein Matrix of Tiere/RenTier\n" << dist2.GetMatrixAsString());
         TRACE("Needed steps: " << dist2.backtrace());
         TRACE("subs: " << dist2.GetSubstCnt() << ", ins :" << dist2.GetInsCnt()
            << " del: " << dist2.GetDelCnt());
         EXPECT_EQ(0, dist2.GetSubstCnt());
         EXPECT_EQ(1, dist2.GetDelCnt());
         EXPECT_EQ(3, dist2.GetInsCnt());
      }

      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }


   TEST(LevenshteinTest, CallsignTest)
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "CallsignTest", "");

      vector<string> recogn{ "swiss", "one", "nine", "mike" };
      vector<string> gold{ "swiss", "one", "nine", "two", "mike" };
      Levenshtein dist(recogn, gold);
      int ld = dist.CalcLevenshteinDistance();
      EXPECT_EQ(1, ld);
      TRACE("Matrix of " << StringVecToOneString(recogn) << "/"
         << StringVecToOneString(gold) << "\n" << dist.GetMatrixAsString());
      EXPECT_EQ(1, dist.GetInsCnt());
      TRACE(dist.GetPrettyPrint(ld, "gold utter", "recogn"));
      TRACE(dist.GetPrettyPrint(ld, "gold", "recognition"));


      vector<string> ABP712{ "bair", "seven", "one", "two" };
      Levenshtein dist2(recogn, ABP712);
      ld = dist2.CalcLevenshteinDistance();
      EXPECT_EQ(4, ld);
      TRACE("Matrix of " << StringVecToOneString(recogn) << "/"
         << StringVecToOneString(ABP712) << "\n" << dist2.GetMatrixAsString());
      TRACE("Needed steps: " << dist2.backtrace());
      TRACE(dist2.GetPrettyPrint(ld, "gold", "recognition"));


      vector<string> ANE2498{ "nostru", "air", "two", "four", "nine", "eight" };
      Levenshtein dist3(recogn, ANE2498);
      ld = dist3.CalcLevenshteinDistance();
      EXPECT_EQ(5, ld);
      TRACE("Matrix of " << StringVecToOneString(recogn) << "/"
         << StringVecToOneString(ANE2498) << "\n" << dist3.GetMatrixAsString());
      TRACE("Needed steps: " << dist3.backtrace());
      TRACE(dist3.GetPrettyPrint(ld, "gold", "recognition"));

      vector<string> SWR129M{ "swiss", "one", "two", "nine", "mike" };
      Levenshtein dist4(gold, SWR129M);
      ld = dist4.CalcLevenshteinDistance();
      EXPECT_EQ(2, ld);
      EXPECT_EQ(1, dist4.GetInsCnt());
      EXPECT_EQ(1, dist4.GetDelCnt());
      TRACE("Matrix of " << StringVecToOneString(gold) << "/"
         << StringVecToOneString(SWR129M) << "\n" << dist4.GetMatrixAsString());
      TRACE("Needed steps: " << dist4.backtrace());
      TRACE(dist4.GetPrettyPrint(ld, "gold", "recognition"));

      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }

   TEST(LevenshteinTest, ThalesExample)
      /* one is much longer than the other, if one is longer than the other
      we need to add in the algorithm the difference to deletion or insertions*/
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "ThalesExample", "");
      string gold = "channex one nine x-ray foxtrot contact praha radar one two zero decimal two seven five";
      string recogn = "channex one niner yankee foxtrot";

      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(12, dist.CalcLevenshteinDistance());
         TRACE("ins: " << dist.GetInsCnt() << " / del: " << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         EXPECT_EQ(0, dist.GetInsCnt());
         EXPECT_EQ(10, dist.GetDelCnt());
         EXPECT_EQ(2, dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(12, dist.CalcLevenshteinDistance());
         TRACE("\nins: " << dist.GetInsCnt() << " / del: " << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         EXPECT_EQ(10, dist.GetInsCnt());
         EXPECT_EQ(0, dist.GetDelCnt());
         EXPECT_EQ(2, dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }

   TEST(LevenshteinTest, ThalesExample2)
      /* one is much longer than the other, if one is longer than the other
      we need to add in the algorithm the difference to deletion or insertions*/
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "ThalesExample2", "");
      string gold = "              nine x-ray foxtrot contact praha radar one two zero decimal two seven five";
      string recogn = "channex one nine x-ray foxtrot                         two zero deci";

      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(10, dist.CalcLevenshteinDistance());
         TRACE("ins: " << dist.GetInsCnt() << " / del: " << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         EXPECT_EQ(2, dist.GetInsCnt());
         EXPECT_EQ(7, dist.GetDelCnt());
         EXPECT_EQ(1, dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(10, dist.CalcLevenshteinDistance());
         TRACE("\nins: " << dist.GetInsCnt() << " / del: " << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         EXPECT_EQ(7, dist.GetInsCnt());
         EXPECT_EQ(2, dist.GetDelCnt());
         EXPECT_EQ(1, dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }

   TEST(LevenshteinTest, ThalesExample3)
      /* real example nearly same size*/
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "ThalesExample3", "");
      string gold = "slovak government     zero zero one dobry den praha radar radar contact descend flight level one two six zero";
      string recogn = " sierra sierra golf zero zero one dobry den praha radar radar contact climb   flight level         six zero ";

      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(6, dist.CalcLevenshteinDistance());
         TRACE("ins: " << dist.GetInsCnt() << " / del: " << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         EXPECT_EQ(1, dist.GetInsCnt());
         EXPECT_EQ(2, dist.GetDelCnt());
         EXPECT_EQ(3, dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(6, dist.CalcLevenshteinDistance());
         TRACE("\nins: " << dist.GetInsCnt() << " / del: " << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         EXPECT_EQ(2, dist.GetInsCnt());
         EXPECT_EQ(1, dist.GetDelCnt());
         EXPECT_EQ(3, dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }



   TEST(LevenshteinTest, differentDistCalc)
      /* we got different results for Levenshtein distance between string1 and string2
        resp. string2 and string1. Now problem is solved  */
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "differentDistCalc", "");
      string gold = "swiss one nine mike";
      string recogn = " swiss air one nine two mike";

      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(2, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(2, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }


   TEST(LevenshteinTest, differentDistCalc2)
      /* we got different results for Levenshtein distance between string1 and string2
        resp. string2 and string1.   */
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "differentDistCalc2", "");
      string gold = "sky travel two four two seven radar contact climb aeh correction descend flight level one hundred";
      string recogn = "sky travel two four two seven radar contact climb     correction descend flight level one hundred";

      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(1, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(1, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }


   TEST(LevenshteinTest, differentDistCalc3)
      /* we got different results for Levenshtein distance between string1 and string2
        resp. string2 and string1.   */
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "differentDistCalc3", "");
      string gold = "egyptian air force four four two  six     climb flight level one four zero";
      string recogn = "sky travel       two four two seven    climb flight level one four zero";
      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(5, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(5, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }


   TEST(LevenshteinTest, differentInsSubstDelValues)
      /* we got more ins/subs/del values then the Levenshtein distance
      for longer one first we get the wrong result:
             dist 3, ins: 0 / del: 2 / subs: 2
      Needed steps: Subs Subs Del Del Equ Equ Equ Equ Equ Equ Equ Equ Equ Equ Equ

      For longer one last, we got the correct result
       dist 3, ins: 2 / del: 0 / subs: 1
      */
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "differentInsSubstDelValues", "");
      string gold = "sky travel five seven   juliett dobry den praha radar radar contact descend flight level one hundred";
      string recogn = "ryan air     juliett dobry den praha radar radar contact descend flight level one hundred";
      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(4, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(goldVect) << "\n          "
            << StringVecToOneString(recogUtterVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }

      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(4, dist.CalcLevenshteinDistance());
         TRACE("\n dist " << dist.CalcLevenshteinDistance() << ", ins: " << dist.GetInsCnt() << " / del: "
            << dist.GetDelCnt() << " / subs: " << dist.GetSubstCnt());
         TRACE("Matrix of " << StringVecToOneString(recogUtterVect) << "\n          "
            << StringVecToOneString(goldVect) << "\n" << dist.GetMatrixAsString()
            << "Needed steps: " << dist.backtrace());
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }

   TEST(LevenshteinTest, CreateInsDelSubsAsString)
      /* we got more ins/subs/del values then the Levenshtein distance
      for longer one first we get the wrong result:
             dist 3, ins: 0 / del: 2 / subs: 2
      Needed steps: Subs Subs Del Del Equ Equ Equ Equ Equ Equ Equ Equ Equ Equ Equ

      For longer one last, we got the correct result
       dist 3, ins: 2 / del: 0 / subs: 1
      */
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "CreateInsDelSubsAsString", "");
      string gold = "sky travel five seven  juliett dobry den praha  radar radar contact descend flight level one hundred";
      string recogn = "ryan air               juliett dobry den prahar radar radar contact descend flight level one hundred";
      vector<string> goldVect = libutil::C_String(gold).split_V(" ", true);
      vector<string> recogUtterVect = libutil::C_String(recogn).split_V(" ", true);
      {
         Levenshtein dist(goldVect, recogUtterVect);
         EXPECT_EQ(5, dist.CalcLevenshteinDistance());
         string op = dist.backtrace();
         TRACE("   " << gold << "\n   " << recogn << "\n op " << op);
      }
      {
         Levenshtein dist(recogUtterVect, goldVect);
         EXPECT_EQ(5, dist.CalcLevenshteinDistance());
         string op = dist.backtrace();
         TRACE("   " << recogn << "\n   " << gold << "\nop " << op);
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }


   TEST(LevenshteinTest, TestLevenshteinDistLong2Helmke)
      // The LD for two strings is calculated the calculation includes
      // The matrix is not big enough to calculate the LD
      // We expect an error and not and LD > 0
      // Correct LD is 3     
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "TestLevenshteinDistLong2Helmke", "");
      vector<string> s1{ "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango2", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x" };
      vector<string> s2{ "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x",
      "Tango", "Tango1", "Tango", "Tango1", "x" , "y" };
      Levenshtein dist(s1, s2);
      int d = dist.CalcLevenshteinDistance();
      if (d > 0 && d != 3)
      {
         cerr << "we expect a crash, but we got " << d << endl;
         return false;
      }
      RETURN_TRUE_IF_NOT_GOOGLE_TEST;
   }








   int Levenshtein::CalcLevenshteinDistance()
      /*
      *  see description at https://de.wikipedia.org/wiki/Levenshtein-Distanz  (16.03.2020)
      *  and C++-code examples at https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance (08.07.2023)
      */
   {
      mi_zCnt_m1 = static_cast<int>(mstr1.size()) + 1;
      mi_spCnt_n2 = static_cast<int>(mstr2.size()) + 1;
      if (mi_zCnt_m1 * mi_zCnt_m1 > MAX_ARRAY_SIZE)
      {
         cerr << "*** two many elements in internal array, we need " <<
            mi_zCnt_m1 * mi_zCnt_m1 << " but have only " << MAX_ARRAY_SIZE
            << " ***" << endl;
         return -1;
      }
      // Init of first elements
      for (int i = 0; i < mi_zCnt_m1; ++i)
      {
         GetMod(i, 0) = i;
      }
      for (int i = 1; i < mi_spCnt_n2; ++i)
      {
         GetMod(0, i) = i;
      }

      for (int i = 1; i < mi_zCnt_m1; ++i)
      {
         for (int j = 1; j < mi_spCnt_n2; ++j)
         {
            if (mstr1[i - 1] == mstr2[j - 1])
            {
               GetMod(i, j) = GetMod(i - 1, j - 1);
            }
            else
            {
               int mini = min(min(GetMod(i - 1, j - 1), // Substitution
                  GetMod(i, j - 1)),  // Insertion
                  GetMod(i - 1, j)  // Deletion
               );
               GetMod(i, j) = mini + 1;
            }
         } // for j
      } // for i

      mb_backtrackCalculated = false;
      return Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1);

   }// Levenshtein::CalcLevenshteinDistance()


   std::string Levenshtein::GetMatrixAsString() const
   {
      stringstream str;
      for (int i = 0; i < mi_zCnt_m1; ++i)
      {
         for (int j = 0; j < mi_spCnt_n2; ++j)
         {
            str << setw(3) << Get(i, j) << " ";
         }
         str << "\n";
      }
      return str.str();
   }// Levenshtein::GetMatrixAsString()


   std::string Levenshtein::backtrace() const
   {
      if (!mb_backtrackCalculated)
      {
         CalcSubsDelIns();
      }
      int stepCnt = 0; // number of word in retValue
      string retValue = backtraceInternal(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1, stepCnt);
      return retValue;
   }

   string Levenshtein::backtraceInternal(int i, int j, int& ai_stepCnt) const
      /* ai_stepCnt incremented with each call */
   {
      if (i > 0 && Get(i - 1, j) + 1 == Get(i, j))
      {
         return backtraceInternal(i - 1, j, ++ai_stepCnt) + "Del ";
      }
      else if (j > 0 && Get(i, j - 1) + 1 == Get(i, j))
      {
         return backtraceInternal(i, j - 1, ++ai_stepCnt) + "Ins ";
      }

      else if (i > 0 && j > 0 && Get(i - 1, j - 1) + 1 == Get(i, j))
      {
         return backtraceInternal(i - 1, j - 1, ++ai_stepCnt) + "Subs ";
      }
      else if (i > 0 && j > 0 && Get(i - 1, j - 1) == Get(i, j))
      {
         return backtraceInternal(i - 1, j - 1, ++ai_stepCnt) + "Equ ";
      }
      else
      {
         return "";
      }
   } // Levenshtein::backtraceInternal



   void Levenshtein::backtraceInternalMember(int i, int j) const
   {
      if (i > 0 && Get(i - 1, j) + 1 == Get(i, j))
      {
         ++mi_delCnt;
         backtraceInternalMember(i - 1, j);
      }
      else if (j > 0 && Get(i, j - 1) + 1 == Get(i, j))
      {
         ++mi_insCnt;
         backtraceInternalMember(i, j - 1);
      }

      else if (i > 0 && j > 0 && Get(i - 1, j - 1) + 1 == Get(i, j))
      {
         ++mi_subsCnt;
         return backtraceInternalMember(i - 1, j - 1);
      }
      else if (i > 0 && j > 0 && Get(i - 1, j - 1) == Get(i, j))
      {
         backtraceInternalMember(i - 1, j - 1);
      }
   } // Levenshtein::backtraceInternalMember





   void Levenshtein::CalcSubsDelIns() const
   {
      LOG_METHOD(g_DEBUGns, g_DEBUGcl, "CalcSubsDelIns", "");

      mi_subsCnt = 0;
      mi_delCnt = 0;
      mi_insCnt = 0;
      Levenshtein* hlp = const_cast<Levenshtein*>(this);
      hlp->CalcLevenshteinDistance();

      backtraceInternalMember(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1);
      mb_backtrackCalculated = true;

      if (Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1) > mi_subsCnt + mi_delCnt + mi_insCnt)
      {
         if (mi_spCnt_n2 < mi_zCnt_m1)
         {
            // we miss some deletions
            TRACE("Correcting by " << Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1) - (mi_subsCnt + mi_delCnt + mi_insCnt)
               << " deletions")
               mi_delCnt += Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1) - (mi_subsCnt + mi_delCnt + mi_insCnt);
         }
         else
         {
            TRACE("Correcting by " << Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1) - (mi_subsCnt + mi_delCnt + mi_insCnt)
               << " insertion")
               mi_insCnt += Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1) - (mi_subsCnt + mi_delCnt + mi_insCnt);
         }
      }
      /*INVARIANT(Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1) ==
         mi_subsCnt + mi_delCnt + mi_insCnt,
         " distance is " << Get(mi_zCnt_m1 - 1, mi_spCnt_n2 - 1)
         << ", but subs " << mi_subsCnt << " del " << mi_delCnt << " ins " << mi_insCnt);*/
   } // Levenshtein::CalcSubsDelIns()


   string Levenshtein::GetPrettyPrint
   (
      int ai_ld,
      string astr_goldText,
      string astr_recognText
   ) const
   {
      std::string prettyGold;
      std::string prettyRecogn;
      vector<string> opsVect = C_String(backtrace()).split_V(" ", true);
      int lenDiff = static_cast<int>(astr_recognText.length() - astr_goldText.length());
      int blankGold = lenDiff < 0 ? 3 : lenDiff + 3;
      int blankRecogn = lenDiff > 0 ? 3 : -lenDiff + 3;

      bool prettyPrintCheck = GoldAndRecogAsPrettyString(
         mstr1, mstr2, opsVect, prettyGold, prettyRecogn);
      ostringstream ostr;
      ostr << astr_goldText << setw(blankGold) << " : ";
      // checking if prettyPrinting has worked else we proceed with normal printing
      if (prettyPrintCheck)
      {
         ostr << prettyGold << "\n";
      }
      else
      {
         ostr << mstr1 << "\n";
      }
      ostr << astr_recognText << setw(blankRecogn) << " : ";
      // checking if prettyPrinting has worked else we proceed with normal printing
      if (prettyPrintCheck)
      {
         ostr << prettyRecogn;
      }
      else
      {
         ostr << mstr2;
      }
      ostr << " // LD: " << ai_ld
         << ", subs: " << GetSubstCnt()
         << " / ins : " << GetInsCnt() << " / del : " << GetDelCnt()
         << " gold words: " << mstr1.size() << "\n";
      return ostr.str();
   }


   bool GoldAndRecogAsPrettyString
   (
      const vector<std::string>& goldVec,
      const vector<std::string>& recognVec,
      const vector<std::string>& opsVec,
      std::string& prettyGold,
      std::string& prettyRecogn
   )
   {
      int goldCounter = 0;
      int recognCounter = 0;
      int goldSize = static_cast<int>(goldVec.size());
      int recognSize = static_cast<int>(recognVec.size());
      for (auto iter : opsVec)
      {
         if (iter == "Subs")
         {
            if (goldCounter < goldSize && recognCounter < recognSize)
            {
               std::string gold = goldVec[goldCounter];
               std::string recogn = recognVec[recognCounter];

               int goldLen = (int)gold.length();
               int recognLen = (int)recogn.length();
               if (goldLen < recognLen)
               {
                  prettyRecogn = prettyRecogn + recogn + " ";
                  prettyGold = prettyGold + gold + std::string(recognLen - goldLen + 1, ' ');
               }
               else
               {
                  prettyGold = prettyGold + gold + " ";
                  prettyRecogn = prettyRecogn + recogn + std::string(goldLen - recognLen + 1, ' ');
               }
               ++goldCounter;
               ++recognCounter;
            } // if (goldCounter < goldSize && recognCounter < recognSize)

            else
               return false;

         } // if (iter == "Subs")

         else if (iter == "Del")
         {
            if (goldCounter < goldSize)
            {
               std::string gold = goldVec[goldCounter];
               int goldLen = (int)gold.length();
               prettyGold = prettyGold + gold + " ";
               prettyRecogn = prettyRecogn + std::string(goldLen + 1, ' ');
               ++goldCounter;
            }

            else
               return false;

         } // else if (iter == "Del")

         else if (iter == "Ins")
         {
            if (recognCounter < recognSize)
            {
               std::string recogn = recognVec[recognCounter];
               int recognLen = (int)recogn.length();
               prettyRecogn = prettyRecogn + recogn + " ";
               prettyGold = prettyGold + std::string(recognLen + 1, ' ');
               ++recognCounter;
            }

            else
               return false;

         } // else if (iter == "Ins")

         else
         {
            if (goldCounter < goldSize && recognCounter < recognSize)
            {
               std::string gold = goldVec[goldCounter];
               std::string recogn = recognVec[recognCounter];
               prettyGold = prettyGold + gold + " ";
               prettyRecogn = prettyRecogn + recogn + " ";
               ++goldCounter;
               ++recognCounter;
            }

            else
               return false;

         } // else
      }

      return true;
   }



